------------------------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local tableComp    = require("scenario/intersection/tableComprehension")
local classHelper  = require("scenario/intersection/classHelper")

------------------------------------------------------------------------------------------------------------------

--[[
NOTE: This class is implied as an ABSTRACT, so no instances should be created!
]]--

local ClassCrashbreaker = classHelper.inheritsFrom(nil)

-------------------------------------------------------------------------------------------------------
-- Static Attributes
----------------------------------------------------------------------------------------------------

ClassCrashbreaker.nextID = 1

----------------------------------------------------------------------------------------------------
-- Static Methods
----------------------------------------------------------------------------------------------------

function ClassCrashbreaker:convert(subclass, instance)
    --[[
    Upgrades a regular crashbreaker to an armageddon version
    Parameters:
        subclass - the new subclass to change the instance to
        instance - the crashbreaker instance
    Returns:
        <subclass instance> - the crashbreaker converted to the new subclass
    ]]--
    if self.isUILinked then
        guihooks.trigger("i_crashbreaker_upgrade", {change = subclass:instanceOf()})
    end

    -- Create new instance and copy across variables
    local newInst = subclass:new(instance:getVehicles(), instance:getRequirement(), instance:UILinked())
    newInst:initEffects()
    return newInst
end

-------------------------------------------------------------------------------------------------------
-- Instance Methods
----------------------------------------------------------------------------------------------------

function ClassCrashbreaker:unlock()
    --[[
    Unlocks the crashbreaker
    ]]--
    self.locked = false
end

function ClassCrashbreaker:setRequirement(req)
    req = req or 0

    errorHandler.assertTrue(req <= tableComp.lengthOfTable(self.vehicles), "vehicleRequirement must be less than the total number of vehicles")
    errorHandler.assertTrue(req >= 0, "vehicleRequirement must be positive")

    self.vehicleRequirement = req

    if req == 0 then
        self.locked = false
    end
end

function ClassCrashbreaker:addCrashedVehicle()
    --[[
    Adds 1 to the vehicles crashed and unlock the crashbreaker if the target is reached
    ]]--
    self.vehiclesCrashed = self.vehiclesCrashed + 1

    if self.locked then
        -- Allow the crashbreaker to be used
        if self.vehiclesCrashed >= self.vehicleRequirement then
            self:unlock()
        end

        -- Don't update when subsequent vehicles after collection are crashed into
        if self.isUILinked and self.vehicleRequirement - self.vehiclesCrashed >= 0 then
            self:updateUI({requirement = self.vehicleRequirement - self.vehiclesCrashed})
        end
    end
end

function ClassCrashbreaker:getRelativeRequirement()
    --[[
    Returns:
        <number> - the number of vehicles left to crash to unlock the crashbreaker
    ]]--
    return math.max(0, self.vehicleRequirement - self.vehiclesCrashed)
end

function ClassCrashbreaker:isUsed()         return self.used               end
function ClassCrashbreaker:isLocked()       return self.locked             end
function ClassCrashbreaker:UILinked()       return self.isUILinked         end
function ClassCrashbreaker:beingUsed()      return self.inUse              end
function ClassCrashbreaker:getVehicles()    return self.vehicles           end
function ClassCrashbreaker:getRequirement() return self.vehicleRequirement end

function ClassCrashbreaker:canBeUsed()
    --[[
    Returns:
        <boolean> - whether the crashbreaker can be used
    ]]--
    return not (self.locked or self.used or self.inUse)
end

function ClassCrashbreaker:reset()
    --[[
    Resets the crashbreaker and the UI to their initial state
    ]]--
    self.vehiclesCrashed = 0
    self.locked = true
    self.used = false
    self.inUse = false

    if self.isUILinked then
        self:updateUI({requirement = self.vehicleRequirement})
    end
end

function ClassCrashbreaker:updateUI(data)
    guihooks.trigger("i_crashbreaker_update", data)
end

-- Abstract method
function ClassCrashbreaker:use()
    error("ClassCrashbreaker:use() - method is abstract and must be implemented by the subclass")
end

function ClassCrashbreaker:instanceOf()
    return "ClassCrashbreaker"
end

------------------------------------------------------------------------------------------------------------------
-- Constructor
----------------------------------------------------------------------------------------------------

function ClassCrashbreaker:new(vehicles, vehicleRequirement, isUILinked)
    --[[
    Parameters:
        vehicles           - the list of vehicle instances
        vehicleRequirement - the amount of vehicles required to unlock the crashbreaker
        isUILinked         - whether this object is linked to the UI (ie. will send messages to the
                             UI)
    ]]--
    local self = {}
    self.vehiclesCrashed = 0
    self.vehicles = vehicles
    self.locked = true
    self.used = false
    self.inUse = false
    self.isUILinked = isUILinked

    self.ID = ClassCrashbreaker.nextID
    ClassCrashbreaker.nextID = ClassCrashbreaker.nextID + 1

    ClassCrashbreaker.setRequirement(self, vehicleRequirement)

    setmetatable(self, {__index = ClassCrashbreaker})

    if self.isUILinked then
        -- Set initial value for UI
        self:updateUI({requirement = self.vehicleRequirement})
    end

    return self
end

------------------------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

return ClassCrashbreaker