local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

-- Classes
local ClassCoin                 = require("scenario/intersection/classes/pickups/classCoin")
local ClassVector               = require("scenario/intersection/classes/classVector")
local ClassPickup               = require("scenario/intersection/classes/pickups/classPickup")
local ClassMultiplier           = require("scenario/intersection/classes/pickups/classMultiplier")
local ClassArmageddon           = require("scenario/intersection/classes/pickups/classArmageddonPickup")
local ClassInstantCrashbreaker  = require("scenario/intersection/classes/pickups/classInstantCrashbreaker")

-- Handlers
local crashbreaker = require("scenario/intersection/handlers/crashbreaker")
local errorHandler = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local pickups = {coins = {}, multipliers = {}, explosives = {}}

----------------------------------------------------------------------------------------------------

local function findPickups()
    --[[
    Finds all pickups in the scenario
    ]]--
    local found = {coins = {}, multipliers = {}, explosives = {}}

    -- Coins
    for i, value in ipairs(ClassCoin:getTypes()) do
        local name = ClassCoin:getPrefix()..value
        local coin = scenetree.findObject(name)
        -- All coins MUST be in the scenario
        errorHandler.assertTrue(coin ~= nil, name.." is missing in this scenario")

        found.coins[i] = {value = value, class = "coin", position = coin:getPosition()}
    end

    -- Multipliers
    for _, value in ipairs(ClassMultiplier:getTypes()) do
        local multi = scenetree.findObject(value)
        -- Not all types of multiplier may be in the scenario
        if multi then
            table.insert(found.multipliers, {
                value = value,
                class = "multiplier",
                position = multi:getPosition()
            })
        end
    end

    -- Armageddon
    local armageddon = scenetree.findObject("armageddon")
    if armageddon then
        table.insert(found.explosives, {
            value = "armageddon",
            class = "armageddon",
            position = armageddon:getPosition()
        })
    end

    -- Instant Crashbreakers
    local i = 1
    local name = "instantCrashbreaker"..i
    local obj = scenetree.findObject(name)

    while obj do
        found.explosives[#found.explosives+1] = {
            value = name,
            class = "instantCrashbreaker",
            position = obj:getPosition()
        }

        i = i + 1
        name = "instantCrashbreaker"..i
        obj = scenetree.findObject(name)
    end

    return found
end

local function createPickupInstances(vehicles)
    --[[
    Creates instances of ClassPickup (ClassPickup is implied as abstract so only those that are
    sub-classes are actually instantiated) from those in-world
    Parameters:
        vehicles - all vehicles which can be affected by instant crashbreakers
    ]]--
    local foundPickups = findPickups()

    for type, found in pairs(foundPickups) do
        for _, data in ipairs(found) do
            local value = data.value
            local class = data.class
            local pos = data.position

            local position = ClassVector:new(pos.x, pos.y, pos.z)

            local instance

            -- Create respective instance
            if class == "coin" then
                instance = ClassCoin:new(position, value)
            elseif class == "multiplier" then
                instance = ClassMultiplier:new(position, value)
            elseif class == "armageddon" then
                instance = ClassArmageddon:new(position, crashbreaker.getCrashbreaker())
            elseif class == "instantCrashbreaker" then
                instance = ClassInstantCrashbreaker:new(position, value, vehicles)
            else
                error((class or "nil").." is not a valid sub-class of ClassPickup")
            end

            pickups[type][value] = instance
        end
    end
end

local function checkCollectables(vehicle)
    --[[
    Checks each pickup to see if this vehicle is close enough to collect them
    If there are, this function also collects them
    Parameters:
      vehicle - the vehicle to check
    ]]--
    errorHandler.assertNil(vehicle)

    for _, v in pairs(pickups) do
        for _, instance in pairs(v) do
            -- No point in checking those that are collected
            if not instance:isCollected() then
                local dist = vehicle:getPosition():getDistanceBetween(instance:getPosition())
                if dist <= ClassPickup.PICKUP_RANGE then
                    if instance:instanceOf("ClassArmageddonPickup") then
                        crashbreaker.setCrashbreaker(instance:collect(crashbreaker.getCrashbreaker()))
                    elseif instance:instanceOf("ClassInstantCrashbreaker") then
                        instance:collect(vehicle)
                    else
                        instance:collect()
                    end
                end
            end
        end
    end
end

local function resetPickups()
    --[[
    Resets the pickups
    ]]--
    -- Simply reset, rather than re-creating
    for _, table in pairs(pickups) do
        for _, instance in pairs(table) do
            instance:reset()
        end
    end
end

local function update(dt)
    --[[
    Updates pickups
    Parameters:
        dt - the time since the last frame
    ]]--
    for _, table in pairs(pickups) do
        for _, instance in pairs(table) do
            instance:update(dt)
        end
    end
end

local function reset()
    --[[
    Resets the crashbreaker and pickups for the next attempt
    ]]--
    resetPickups()
end

local function resetState()
    --[[
    Resets variables back to their initial state
    ]]--
    pickups = {coins = {}, multipliers = {}, explosives = {}}
end

----------------------------------------------------------------------------------------------------
-- Getters
----------------------------------------------------------------------------------------------------

local function getPickups()      return pickups      end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.createPickupInstances = createPickupInstances
M.checkCollectables = checkCollectables
M.update = update
M.reset = reset
M.getPickups = getPickups
M.resetState = resetState
return M